#include "st.h"

/* Create an uninitialized image. */
STHDRImage::STHDRImage() 
:mWidth(0), mHeight(0), mPixels(NULL)
{
}

/* Create an HDR image w x h in dimension and clear all values to 0. */
STHDRImage::STHDRImage(int w, int h) {
	mWidth = w;
	mHeight = h;
	mPixels = new STColor[w*h];
	memset(mPixels, 0, w*h*sizeof(STColor));
}
/* Destroy the HDR image. */
STHDRImage::~STHDRImage() {
	if (mPixels)
		delete[] mPixels;
}

int 
STHDRImage::GetHeight() const
{
	return mHeight; 
}
int 
STHDRImage::GetWidth() const {
	return mWidth; 
}

/* Resize the hdr image to the specified size
 * and zero out its contents.
 */
void 
STHDRImage::Resize(int nw, int nh) {
	if (mPixels)
		delete[] mPixels;
	mWidth = nw;
	mHeight = nh;
	mPixels = new STColor[nw*nh];
	memset(mPixels, 0, nw*nh*sizeof(STColor));
}

/* Pixel accessor, pixels in HDR images are floating point colors. */
STColor& 
STHDRImage::operator()(int x, int y) {
	return mPixels[y*mWidth+x]; 
}
/* Get a pointer to the raw pixel data. */
STColor* 
STHDRImage::Pixels() { 
	return mPixels; 
}

/**
* Loads an image off disk.  This method does not do any real work, it
* just calls a format specific method based on the image format (format
* determined by file extension)
*/ 
STStatus 
STHDRImage::Load(const char* filename) {
	// Determine how to load the file from the file's extension
	// The ST library currently only supports hdr images of type PFM

	if (filename == NULL)
		return ST_ERROR;

	std::string ext = getExtension( filename );

	if (ext.compare("PFM") == 0)
		return LoadPFM(filename);
	else {
		fprintf(stderr, "Unknown image file type.\n");
		return ST_ERROR;
	} 
}

/* Load image data from file in PFM format */
STStatus
STHDRImage::LoadPFM(const char* filename) {
	FILE* fin = fopen(filename, "rb");
	char buffer[128];

	if (fin == NULL) {
		printf("Error opening file: %s\n", filename);
		return ST_ERROR;
	}

	// Check the format
	char format;
	fgets(buffer, 80, fin);
	if ( sscanf(buffer, "P%c", &format) != 1 || (format != 'f' && format != 'F') ) {
		printf("Invalid PFM file\n");
		fclose(fin);
		return ST_ERROR;
	}
	// Get the dimensions
	int w, h;
	fgets(buffer, 80, fin);
	if ( sscanf(buffer, "%i %i", &w, &h) != 2 || w <= 0 || h <= 0 ) {
		printf("Invalid PFM file\n");
		fclose(fin);
		return ST_ERROR;
	}
	// Get the scale factor, we just discard this
	// FIXME endianness
	float scale;
	fgets(buffer, 80, fin);
	if ( sscanf(buffer, "%f", &scale) != 1 ) {
		printf("Invalid PFM file\n");
		fclose(fin);
		return ST_ERROR;
	}

	// Create space
	if (mPixels != NULL) delete[] mPixels;
	mWidth = w;
	mHeight = h;
	mPixels = new STColor[w*h];

	// Now we can read the data
	if (format == 'f') { // grayscale
		float gval;
		for(int j = 0; j < h; j++) {
			for(int i = 0; i < w; i++) {
				if ( fread(&gval, sizeof(float), 1, fin) != 1 ) {
					printf("Invalid PFM file\n");
					fclose(fin);
					return ST_ERROR;
				}
				(*this)(i,j) = STColor(gval, gval, gval);
			}
		}
	}
	else { // if (format == 'F') color
		if ( fread(mPixels, sizeof(STColor), w*h, fin) != w*h ) {
			printf("Invalid PFM file\n");
			fclose(fin);
			return ST_ERROR;
		}
	}

	fclose(fin);

	return ST_OK;
}

/**
* Saves an image to disk.  This method does not do any real work, it
* just calls a format specific method based on the image format
* (format determined by file extension)
*/ 
STStatus 
STHDRImage::Save(const char* filename) {
	if (filename == NULL || mPixels == NULL)
		return ST_ERROR;

	std::string ext = getExtension( filename );

	if (ext.compare("PFM") == 0)
		return SavePFM(filename);
	else {
		fprintf(stderr, "Unknown image file type.\n");
		return ST_ERROR;
	} 
}

/* Writes image information in PFM format. */
STStatus 
STHDRImage::SavePFM(const char* filename) {
	FILE* fout = fopen(filename, "wb");
	if (fout == NULL) {
		printf("Error opening file: %s\n", filename);
		return ST_ERROR;
	}
	fprintf(fout, "PF\n");
	fprintf(fout, "%i %i\n", mWidth, mHeight);
	fprintf(fout, "-1\n"); // FIXME endianness
	fwrite(mPixels, sizeof(STColor), mWidth*mHeight, fout);
	fclose(fout);

	return ST_OK;
}

