
#include "stgl.h"
#include "STText.h"


/**
* Constructor accepts a STFontFace and a string as input. A local copy
* of the string is made, but only a reference to the face object is
* maintained in the STText class.  STText does not own the face object.
*/
STText::STText(STFontFace* fontFace, const char* text) :
	mFontFace(fontFace), mText(text), mColor(1.0f, 1.0f, 1.0f)
{

}

/**
*	Returns the string of text rendered in this text box
*/
const char*
STText::GetText() const {
	return mText.c_str();
}


/**
* Sets contents of the text box to contain the specified string
*/
void
STText::SetText(const char* text) {
	mText = std::string(text);
}


/**
* The width of the text box in units of *PIXELS*.  This is a floating
* point value to permit subpixel precision
*/
float
STText::GetWidth() {
	return mFontFace->ComputeWidth(mText.c_str());
}


/**
* Returns a pointer to the font face object used by this text box to
* render the text
*/
STFontFace*
STText::GetFontFace() {
	return mFontFace;
}


/**
* Changes the font face used to render text
*/
void
STText::SetFontFace(STFontFace* face) {
	mFontFace = face;
}

/**
* Returns the color of the text
*/
const STColor&
STText::GetColor() const {
	return mColor;
}


/**
* Sets the color of text in this text box to the color 'c'
*/
void
STText::SetColor(const STColor& c) {
	mColor = c;
}


/**
* Draw the text to the screen using the text box's assigned font face
*/
void
STText::Draw() {

	// Need to set some OpenGL state up front.  In order to be able to resotre the old
	// state at the end of the call, we'll push a new entry onto the client and regular
	// attribute stacks.  We'll pop the stacks at the end of the cdall to return to
	// the previous GL configuration
	glPushClientAttrib(GL_CLIENT_PIXEL_STORE_BIT);
    glPushAttrib(GL_ENABLE_BIT);

	glPixelStorei(GL_UNPACK_LSB_FIRST, GL_FALSE);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glDisable(GL_BLEND);

	// text always begins at (0.0f, 0,0f) in the local space.
	glColor3f(mColor.r, mColor.g, mColor.b);
	glRasterPos2f(0.0f, 0.0f);

	// all the heavy lifting happens here, see STFontFace::DrawString()
	float width = mFontFace->DrawString(mText.c_str());

	// restore GL state to what is was at the beginning of the call
	glPopAttrib();
    glPopClientAttrib();

	// The follow code is useful for debugging.  It draws a red outline
	// around the text box.
	/*
	glColor3f(1.0f, 0.0f, 0.0f);
	glBegin(GL_LINE_LOOP);
	glVertex2f(0.0f, mFontFace->GetDescender());
	glVertex2f(width, mFontFace->GetDescender());
	glVertex2f(width, mFontFace->GetAscender());
	glVertex2f(0.0f, mFontFace->GetAscender());
	glEnd();
	*/
	
}

